from typing import List, Optional

from apipeline.frames.sys_frames import Frame
from apipeline.processors.frame_processor import FrameDirection, FrameProcessor


class FrameFilter(FrameProcessor):

    def __init__(self,
                 ignored_frame_types: Optional[list] = [],
                 include_frame_types: Optional[list] = None):
        super().__init__()
        self._ignored_frame_types = tuple(
            ignored_frame_types) if ignored_frame_types is not None else None
        self._include_frame_types = tuple(include_frame_types) if include_frame_types else None

    def _should_passthrough_frame(self, frame):
        if self._ignored_frame_types is not None \
                and not isinstance(frame, self._ignored_frame_types):
            if not self._include_frame_types \
                or (self._include_frame_types
                    and isinstance(frame, self._include_frame_types)):
                return False
        return True

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        if self._should_passthrough_frame(frame):
            await self.push_frame(frame, direction)




import asyncio
from apipeline.frames.control_frames import EndFrame
from apipeline.frames.data_frames import TextFrame
from apipeline.pipeline.pipeline import Pipeline
from apipeline.pipeline.runner import PipelineRunner
from apipeline.pipeline.task import PipelineParams, PipelineTask
from apipeline.processors.aggregators.sentence import SentenceAggregator
from apipeline.processors.output_processor import OutputFrameProcessor


async def main():
    aggregator = SentenceAggregator()
    out_processor = OutputFrameProcessor(
        cb=lambda x: print(f"sink_callback print frame: {x}"))
    pipeline = Pipeline([aggregator, out_processor])
    task = PipelineTask(pipeline, PipelineParams(allow_interruptions=True))

    await task.queue_frame(TextFrame("你好，hello. hi"))
    await task.queue_frame(TextFrame("你叫什么？"))
    await task.queue_frame(TextFrame("Hello, "))
    await task.queue_frame(TextFrame("world."))
    await task.queue_frame(TextFrame("hi"))
    await task.queue_frame(EndFrame())

    runner = PipelineRunner()
    await runner.run(task)


if __name__ == "__main__":
    asyncio.run(main())



#
# Copyright (c) 2024, Daily
#
# SPDX-License-Identifier: BSD 2-Clause License
#

from itertools import chain
from typing import List
import asyncio
import logging

from apipeline.frames.control_frames import EndFrame, SyncFrame
from apipeline.frames.sys_frames import SystemFrame
from apipeline.pipeline.base_pipeline import BasePipeline
from apipeline.pipeline.pipeline import Pipeline
from apipeline.processors.frame_processor import FrameDirection, FrameProcessor
from apipeline.frames.base import Frame


class Source(FrameProcessor):

    def __init__(self, upstream_queue: asyncio.Queue):
        super().__init__()
        self._up_queue = upstream_queue

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        match direction:
            case FrameDirection.UPSTREAM:
                await self._up_queue.put(frame)
            case FrameDirection.DOWNSTREAM:
                await self.push_frame(frame, direction)


class Sink(FrameProcessor):

    def __init__(self, downstream_queue: asyncio.Queue):
        super().__init__()
        self._down_queue = downstream_queue

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        match direction:
            case FrameDirection.UPSTREAM:
                await self.push_frame(frame, direction)
            case FrameDirection.DOWNSTREAM:
                await self._down_queue.put(frame)


class SyncParallelPipeline(BasePipeline):
    def __init__(self, *args):
        super().__init__()

        if len(args) == 0:
            raise Exception(f"ParallelTask needs at least one argument")

        self._sinks = []
        self._sources = []
        self._pipelines = []

        self._up_queue = asyncio.Queue()
        self._down_queue = asyncio.Queue()

        logging.debug(f"Creating {self} pipelines")
        for processors in args:
            if not isinstance(processors, list):
                raise TypeError(f"ParallelTask argument {processors} is not a list")

            # We add a source at the beginning of the pipeline and a sink at the end.
            up_queue = asyncio.Queue()
            down_queue = asyncio.Queue()
            source = Source(up_queue)
            sink = Sink(down_queue)
            processors: List[FrameProcessor] = [source] + processors + [sink]

            # Keep track of sources and sinks. We also keep the output queue of
            # the source and the sinks so we can use it later.
            self._sources.append({"processor": source, "queue": down_queue})
            self._sinks.append({"processor": sink, "queue": up_queue})

            # Create pipeline
            pipeline = Pipeline(processors)
            self._pipelines.append(pipeline)
        logging.debug(f"Finished creating {self} pipelines")

    #
    # BasePipeline
    #

    def processors_with_metrics(self) -> List[FrameProcessor]:
        return list(chain.from_iterable(p.processors_with_metrics() for p in self._pipelines))

    #
    # Frame processor
    #

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        # The last processor of each pipeline needs to be synchronous otherwise
        # this element won't work. Since, we know it should be synchronous we
        # push a SyncFrame. Since frames are ordered we know this frame will be
        # pushed after the synchronous processor has pushed its data allowing us
        # to synchrnonize all the internal pipelines by waiting for the
        # SyncFrame in all of them.
        async def wait_for_sync(
            obj, main_queue: asyncio.Queue, frame: Frame, direction: FrameDirection
        ):
            processor: FrameProcessor = obj["processor"]
            queue: asyncio.Queue = obj["queue"]
            await processor.process_frame(frame, direction)
            if isinstance(frame, (SystemFrame, EndFrame)):
                new_frame = await queue.get()
                if isinstance(new_frame, (SystemFrame, EndFrame)):
                    await main_queue.put(new_frame)
                else:
                    while not isinstance(new_frame, (SystemFrame, EndFrame)):
                        await main_queue.put(new_frame)
                        queue.task_done()
                        new_frame = await queue.get()
            else:
                await processor.process_frame(SyncFrame(), direction)
                new_frame = await queue.get()
                while not isinstance(new_frame, SyncFrame):
                    await main_queue.put(new_frame)
                    queue.task_done()
                    new_frame = await queue.get()

        if direction == FrameDirection.UPSTREAM:
            # If we get an upstream frame we process it in each sink.
            await asyncio.gather(
                *[wait_for_sync(s, self._up_queue, frame, direction) for s in self._sinks]
            )
        elif direction == FrameDirection.DOWNSTREAM:
            # If we get a downstream frame we process it in each source
            await asyncio.gather(
                *[wait_for_sync(s, self._down_queue, frame, direction) for s in self._sources]
            )

        seen_ids = set()
        while not self._up_queue.empty():
            frame = await self._up_queue.get()
            if frame and frame.id not in seen_ids:
                await self.push_frame(frame, FrameDirection.UPSTREAM)
                seen_ids.add(frame.id)
            self._up_queue.task_done()

        seen_ids = set()
        while not self._down_queue.empty():
            frame = await self._down_queue.get()
            if frame and frame.id not in seen_ids:
                await self.push_frame(frame, FrameDirection.DOWNSTREAM)
                seen_ids.add(frame.id)
            self._down_queue.task_done()



import unittest

from apipeline.frames.data_frames import AudioRawFrame, TextFrame, ImageRawFrame
from apipeline.serializers.protobuf import ProtobufFrameSerializer


"""
python -m unittest tests.serializer.test_protobuf_serializer.TestProtobufFrameSerializer.test_text
python -m unittest tests.serializer.test_protobuf_serializer.TestProtobufFrameSerializer.test_audio
python -m unittest tests.serializer.test_protobuf_serializer.TestProtobufFrameSerializer.test_image
"""


class TestProtobufFrameSerializer(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.serializer = ProtobufFrameSerializer()

    @unittest.skip("FIXED")
    async def test_skip(self):
        pass

    async def test_text(self):
        text_frame = TextFrame(text="hello world")
        frame = self.serializer.deserialize(self.serializer.serialize(text_frame))
        print(frame)
        self.assertEqual(frame, text_frame)

    async def test_audio(self):
        audio_frame = AudioRawFrame(audio=b"1234567890")
        frame = self.serializer.deserialize(self.serializer.serialize(audio_frame))
        print(frame)
        self.assertEqual(frame, audio_frame)

    async def test_image(self):
        image_frame = ImageRawFrame(
            image=b"1234567890", size=(1280, 720), format="JPEG", mode="RGB",
        )
        frame = self.serializer.deserialize(self.serializer.serialize(image_frame))
        print(frame)
        self.assertEqual(frame, image_frame)



import logging

from asyncio import AbstractEventLoop
import asyncio
import unittest
from apipeline.frames.control_frames import EndFrame
from apipeline.pipeline.pipeline import Pipeline, FrameDirection
from apipeline.pipeline.sync_parallel_pipeline import SyncParallelPipeline
from apipeline.pipeline.task import PipelineTask, PipelineParams
from apipeline.pipeline.runner import PipelineRunner
from apipeline.frames.data_frames import Frame, TextFrame
from apipeline.processors.frame_processor import FrameProcessor


"""
python -m unittest tests.pipeline.test_sync_parallel_pipeline.TestParallelPipeline
"""


class FrameTraceLogger(FrameProcessor):

    def __init__(
            self,
            tag: str,
            *,
            name: str | None = None,
            loop: AbstractEventLoop | None = None,
            **kwargs):
        super().__init__(name=name, loop=loop, **kwargs)
        self._tag = tag

    async def process_frame(
            self,
            frame: Frame,
            direction: FrameDirection = FrameDirection.DOWNSTREAM):
        await super().process_frame(frame, direction)

        from_to = f"{self._prev} ---> {self}"
        if direction == FrameDirection.UPSTREAM:
            from_to = f"{self} <--- {self._next} "
        if self._tag == "1.0":
            await asyncio.sleep(1)
        logging.info(f"Tag: {self._tag}; {from_to} get Frame: {frame}")
        await self.push_frame(frame, direction)


class TestParallelPipeline(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        logging.basicConfig(level=logging.INFO, format="%(asctime)s %(levelname)s %(message)s")

        pipeline = Pipeline([
            FrameTraceLogger(tag="0"),
            SyncParallelPipeline(
                [FrameTraceLogger(tag="1.0"), FrameTraceLogger(tag="1.1")],
                [FrameTraceLogger(tag="2.0"), FrameTraceLogger(tag="2.1")],
            ),
            FrameTraceLogger(tag="3"),
        ])

        self.task = PipelineTask(
            pipeline,
            PipelineParams()
        )

    async def asyncTearDown(self):
        pass

    async def test_run(self):
        runner = PipelineRunner()
        await self.task.queue_frame(TextFrame("你好"))
        await self.task.queue_frame(EndFrame())
        await runner.run(self.task)



#
# Copyright (c) 2024, Daily
#
# SPDX-License-Identifier: BSD 2-Clause License
#

from abc import abstractmethod

from typing import List

from apipeline.processors.frame_processor import FrameProcessor


class BasePipeline(FrameProcessor):

    def __init__(self):
        super().__init__()

    @abstractmethod
    def processors_with_metrics(self) -> List[FrameProcessor]:
        pass



#
# Copyright (c) 2024, Daily
#
# SPDX-License-Identifier: BSD 2-Clause License
#

from dataclasses import dataclass, field

from apipeline.utils.obj import obj_count, obj_id


@dataclass
class Frame:
    id: int = field(init=False)
    name: str = field(init=False)

    def __post_init__(self):
        self.id: int = obj_id()
        self.name: str = f"{self.__class__.__name__}#{obj_count(self)}"

    def __str__(self):
        return self.name



#
# Copyright (c) 2024, Daily
#
# SPDX-License-Identifier: BSD 2-Clause License
#

import logging
import asyncio

from apipeline.frames.sys_frames import Frame, StartInterruptionFrame
from apipeline.frames.control_frames import EndFrame
from apipeline.processors.frame_processor import FrameDirection, FrameProcessor


class AsyncFrameProcessor(FrameProcessor):

    def __init__(
            self,
            *,
            name: str | None = None,
            loop: asyncio.AbstractEventLoop | None = None,
            **kwargs):
        super().__init__(name=name, loop=loop, **kwargs)

        self._push_frame_task = None
        # Create push frame task. This is the task that will push frames in
        # order. We also guarantee that all frames are pushed in the same task.
        self._create_push_task()

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        if isinstance(frame, StartInterruptionFrame):
            await self._handle_interruptions(frame)

    async def cleanup(self):
        if self._push_frame_task:
            self._push_frame_task.cancel()
            await self._push_frame_task

    #
    # Handle interruptions
    #
    async def _handle_interruptions(self, frame: Frame):
        # Cancel the task. This will stop pushing frames downstream.
        self._push_frame_task.cancel()
        await self._push_frame_task
        self._push_frame_task = None
        # Push an out-of-band frame (i.e. not using the ordered push
        # frame task).
        await self.push_frame(frame)
        # Create a new queue and task.
        self._create_push_task()

    #
    # Push frames task
    #

    def _create_push_task(self):
        """
        NOTE: all async frame processors must have create a new queue and task, don't to check task is None
        """
        # logging.debug(f"{self} create push task {self._push_frame_task}")
        self._push_queue = asyncio.Queue()
        self._push_frame_task = self.get_event_loop().create_task(self._push_frame_task_handler())

    async def queue_frame(
            self,
            frame: Frame,
            direction: FrameDirection = FrameDirection.DOWNSTREAM):
        await self._push_queue.put((frame, direction))

    async def _push_frame_task_handler(self):
        running = True
        while running:
            try:
                (frame, direction) = await asyncio.wait_for(self._push_queue.get(), timeout=1)
                await self.push_frame(frame, direction)
                running = not isinstance(frame, EndFrame)
            except asyncio.TimeoutError:
                pass
            except asyncio.CancelledError:
                logging.info(f"{self.name} _push_frame_task_handle cancelled")
                break
            except Exception as ex:
                logging.exception(f"{self.name} Unexpected error in _push_frame_task_handler: {ex}")
                if self.get_event_loop().is_closed():
                    logging.warning(f"{self.name} event loop is closed")
                    break



#
# Copyright (c) 2024, Daily
#
# SPDX-License-Identifier: BSD 2-Clause License
#

from abc import ABC, abstractmethod


class BaseNotifier(ABC):
    @abstractmethod
    async def notify(self):
        pass

    @abstractmethod
    async def wait(self):
        pass



import logging
import dataclasses

import apipeline.frames.protobufs.data_frames_pb2 as frame_protos

from apipeline.frames.data_frames import AudioRawFrame, Frame, TextFrame, ImageRawFrame
from .base_serializer import FrameSerializer


class ProtobufFrameSerializer(FrameSerializer):
    SERIALIZABLE_TYPES = {
        TextFrame: "text",
        AudioRawFrame: "audio",
        ImageRawFrame: "image",
    }

    SERIALIZABLE_FIELDS = {v: k for k, v in SERIALIZABLE_TYPES.items()}

    def serialize(self, frame: Frame) -> str | bytes | None:
        proto_frame = frame_protos.Frame()
        if type(frame) not in ProtobufFrameSerializer.SERIALIZABLE_TYPES:
            logging.warning(f"Frame type {type(frame)} is not serializable")
            return None

        # ignoring linter errors; we check that type(frame) is in this dict above
        proto_optional_name = ProtobufFrameSerializer.SERIALIZABLE_TYPES[type(frame)]
        for field in dataclasses.fields(frame):
            value = getattr(frame, field.name)
            if not value:
                continue

            if (isinstance(frame, ImageRawFrame) and field.name == "size"
                    and isinstance(value, tuple) and len(value) == 2):
                setattr(
                    getattr(
                        proto_frame,
                        proto_optional_name),
                    field.name,
                    f"{value[0]}x{value[1]}")
            else:
                setattr(getattr(proto_frame, proto_optional_name), field.name, value)

        result = proto_frame.SerializeToString()
        return result

    def deserialize(self, data: str | bytes) -> Frame | None:
        """Returns a Frame object from a Frame protobuf. Used to convert frames
        passed over the wire as protobufs to Frame objects used in pipelines
        and frame processors.

        >>> serializer = ProtobufFrameSerializer()
        >>> serializer.deserialize(
        ...     serializer.serialize(OutputAudioFrame(data=b'1234567890')))
        OutputAudioFrame(data=b'1234567890')

        >>> serializer.deserialize(
        ...     serializer.serialize(TextFrame(text='hello world')))
        TextFrame(text='hello world')

        """

        proto = frame_protos.Frame.FromString(data)
        which = proto.WhichOneof("frame")
        if which not in ProtobufFrameSerializer.SERIALIZABLE_FIELDS:
            logging.error("Unable to deserialize a valid frame")
            return None

        args = getattr(proto, which)
        args_dict = {}
        for field in proto.DESCRIPTOR.fields_by_name[which].message_type.fields:
            args_dict[field.name] = getattr(args, field.name)

        # Remove id name
        if "id" in args_dict:
            del args_dict["id"]
        if "name" in args_dict:
            del args_dict["name"]

        # Create the instance
        class_name = ProtobufFrameSerializer.SERIALIZABLE_FIELDS[which]
        if class_name == ImageRawFrame:
            size = args_dict["size"].split("x") if "size" in args_dict else []
            args_dict["size"] = (int(size[0]), int(size[1]))
        instance = class_name(**args_dict)

        # Set Frame id name
        if hasattr(args, "id"):
            setattr(instance, "id", getattr(args, "id"))
        if hasattr(args, "name"):
            setattr(instance, "name", getattr(args, "name"))

        return instance



from typing import Any, Mapping, List
from dataclasses import dataclass


from .base import Frame

#
# System frames
#


@dataclass
class SystemFrame(Frame):
    pass


@dataclass
class CancelFrame(SystemFrame):
    """Indicates that a pipeline needs to stop right away."""
    pass


@dataclass
class ErrorFrame(SystemFrame):
    """This is used notify upstream that an error has occurred downstream the
    pipeline."""
    error: str
    fatal: bool = False

    def __str__(self):
        return f"{self.name}(error: {self.error}, fatal: {self.fatal})"


@dataclass
class StopTaskFrame(SystemFrame):
    """Indicates that a pipeline task should be stopped but that the pipeline
    processors should be kept in a running state. This is normally queued from
    the pipeline task.

    """
    pass


@dataclass
class StartInterruptionFrame(SystemFrame):
    """e.g. Emitted by VAD to indicate that a user has started speaking (i.e. is
    interruption). This is similar to UserStartedSpeakingFrame except that it
    should be pushed concurrently with other frames (so the order is not
    guaranteed).

    """
    pass


@dataclass
class StopInterruptionFrame(SystemFrame):
    """e.g. Emitted by VAD to indicate that a user has stopped speaking (i.e. no more
    interruptions). This is similar to UserStoppedSpeakingFrame except that it
    should be pushed concurrently with other frames (so the order is not
    guaranteed).

    """
    pass


@dataclass
class MetricsFrame(SystemFrame):
    """Emitted by processor that can compute metrics like latencies.
    """
    ttfb: List[Mapping[str, Any]] | None = None
    processing: List[Mapping[str, Any]] | None = None
    tokens: List[Mapping[str, Any]] | None = None
    characters: List[Mapping[str, Any]] | None = None

    def __str__(self):
        p_str = f"{self.name} ttfb:{self.ttfb} | processing:{self.processing} | tokens:{self.tokens} | characters:{self.characters}"
        return p_str



import logging
import asyncio

import unittest

from apipeline.frames.control_frames import EndFrame
from apipeline.frames.data_frames import DataFrame, TextFrame
from apipeline.pipeline.pipeline import Pipeline
from apipeline.pipeline.runner import PipelineRunner
from apipeline.pipeline.task import PipelineParams, PipelineTask
from apipeline.processors.aggregators.sentence import SentenceAggregator
from apipeline.processors.frame_processor import FrameProcessor
from apipeline.processors.text_transformer import StatelessTextTransformer
from apipeline.processors.input_processor import InputFrameProcessor
from apipeline.processors.output_processor import OutputFrameProcessor


"""
python -m unittest tests.pipeline.test_pipeline.TestSentenceAggregatorPipeline.test_pipeline_simple
python -m unittest tests.pipeline.test_pipeline.TestSentenceAggregatorPipeline.test_pipeline_multiple_stages
"""


class TestSentenceAggregatorPipeline(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls):
        logging.basicConfig(level=logging.DEBUG)
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    async def asyncSetUp(self):
        pass

    async def asyncTearDown(self):
        pass

    async def test_pipeline_simple(self):
        aggregator = SentenceAggregator()

        self.text = []

        async def sink_callback(frame: DataFrame):
            print(f"sink_callback print frame: {frame}")
            self.text.append(frame.text)
            # note: if assert false, wait for a while
            # self.assertEqual(frame.text, 'Hello, world.')
        out_processor = OutputFrameProcessor(cb=sink_callback)
        # out_processor = OutputFrameProcessor(cb=lambda x: print(f"sink_callback print frame: {x}"))

        pipeline = Pipeline([aggregator, out_processor])
        task = PipelineTask(pipeline, PipelineParams(allow_interruptions=True))

        await task.queue_frame(TextFrame("Hello, "))
        await task.queue_frame(TextFrame("world."))
        await task.queue_frame(TextFrame("hi"))
        await task.queue_frame(EndFrame())

        runner = PipelineRunner()

        await asyncio.gather(runner.run(task))
        self.assertEqual(self.text, ["Hello, world.", "hi"])

    async def test_pipeline_multiple_stages(self):
        sentence_aggregator = SentenceAggregator()

        def to_upper(x: str):
            return x.upper()
        to_upper_processor = StatelessTextTransformer(to_upper)
        add_space_processor = StatelessTextTransformer(lambda x: x + " ")

        self.text = []

        async def sink_callback(frame: DataFrame):
            print(f"sink_callback print frame: {frame}")
            self.text.append(frame.text)
        out_processor = OutputFrameProcessor(cb=sink_callback)

        pipeline = Pipeline([
            add_space_processor,
            sentence_aggregator,
            to_upper_processor,
            out_processor,
        ])
        task = PipelineTask(pipeline, PipelineParams(allow_interruptions=True))

        sentence = "Hello, world. It's me, a pipeline."
        for c in sentence:
            await task.queue_frame(TextFrame(c))
        await task.queue_frame(EndFrame())

        runner = PipelineRunner()

        await asyncio.gather(runner.run(task))

        print(self.text)
        self.assertEqual(self.text, [
            "H E L L O ,   W O R L D . ",
            "  I T ' S   M E ,   A   P I P E L I N E . ",
        ])
